@@ -4,7 +4,7 @@
 
 ;; Author: Vasilij Schneidermann <mail@vasilij.de>
 ;; URL: https://github.com/wasamasa/nov.el
-;; Version: 0.2.9
+;; Version: 0.3.0
 ;; Package-Requires: ((dash "2.12.0") (esxml "0.3.3") (emacs "24.4"))
 ;; Keywords: hypermedia, multimedia, epub
 
@@ -32,9 +32,11 @@
 ;; - Basic navigation (jump to TOC, previous/next chapter)
 ;; - Remembering and restoring the last read position
 ;; - Jump to next chapter when scrolling beyond end
+;; - Storing and following Org links to EPUB files
 ;; - Renders EPUB2 (.ncx) and EPUB3 (<nav>) TOCs
 ;; - Hyperlinks to internal and external targets
 ;; - Supports textual and image documents
+;; - Info-style history navigation
 ;; - View source of document files
 ;; - Metadata display
 ;; - Image rescaling
@@ -46,6 +48,9 @@
 (require 'esxml-query)
 (require 'shr)
 (require 'url-parse)
+(require 'recentf)
+(require 'bookmark)
+(require 'org)
 
 (when (not (fboundp 'libxml-parse-xml-region))
   (message "Your Emacs wasn't compiled with libxml support"))
@@ -71,10 +76,11 @@ Otherwise the default face is used."
 (defcustom nov-text-width nil
   "Width filled text shall occupy.
 An integer is interpreted as the number of columns.  If nil, use
-the full window's width.  Note that this variable only has an
-effect in Emacs 25.1 or greater."
+the full window's width.  If t, disable filling completely.  Note
+that this variable only has an effect in Emacs 25.1 or greater."
   :type '(choice (integer :tag "Fixed width in characters")
-                 (const   :tag "Use the width of the window" nil))
+                 (const   :tag "Use the width of the window" nil)
+                 (const   :tag "Disable filling" t))
   :group 'nov)
 
 (defcustom nov-render-html-function 'nov-render-html
@@ -126,6 +132,14 @@ Each alist item consists of the identifier and full path.")
 (defvar-local nov-toc-id nil
   "TOC identifier of the EPUB buffer.")
 
+(defvar-local nov-history nil
+  "Stack of documents user has visited.
+Each element of the stack is a list (NODEINDEX BUFFERPOS).")
+
+(defvar-local nov-history-forward nil
+  "Stack of documents user has visited with `nov-history-back' command.
+Each element of the stack is a list (NODEINDEX BUFFERPOS).")
+
 (defun nov-make-path (directory file)
   "Create a path from DIRECTORY and FILE."
   (concat (file-name-as-directory directory) file))
@@ -380,7 +394,10 @@ Each alist item consists of the identifier and full path."
     (define-key map (kbd "p") 'nov-previous-document)
     (define-key map (kbd "[") 'nov-previous-document)
     (define-key map (kbd "t") 'nov-goto-toc)
+    (define-key map (kbd "l") 'nov-history-back)
+    (define-key map (kbd "r") 'nov-history-forward)
     (define-key map (kbd "RET") 'nov-browse-url)
+    (define-key map (kbd "c") 'nov-copy-url)
     (define-key map (kbd "<follow-link>") 'mouse-face)
     (define-key map (kbd "<mouse-2>") 'nov-browse-url)
     (define-key map (kbd "TAB") 'shr-next-link)
@@ -426,6 +443,7 @@ This function honors `shr-max-image-proportion' if possible."
   (cond
    ((not (display-graphic-p))
     (insert alt))
+   ;; TODO: add native resizing support once it's official
    ((fboundp 'imagemagick-types)
     ;; adapted from `shr-rescale-image'
     (let ((edges (window-inside-pixel-edges
@@ -489,12 +507,15 @@ chapter title."
   (let (;; HACK: make buttons use our own commands
         (shr-map nov-mode-map)
         (shr-external-rendering-functions nov-shr-rendering-functions)
-        (shr-use-fonts nov-variable-pitch)
-        (shr-width nov-text-width))
+        (shr-use-fonts nov-variable-pitch))
     ;; HACK: `shr-external-rendering-functions' doesn't cover
     ;; every usage of `shr-tag-img'
     (cl-letf (((symbol-function 'shr-tag-img) 'nov-render-img))
-      (shr-render-region (point-min) (point-max))))
+      (if (eq nov-text-width t)
+          (cl-letf (((symbol-function 'shr-fill-line) 'ignore))
+            (shr-render-region (point-min) (point-max)))
+        (let ((shr-width nov-text-width))
+          (shr-render-region (point-min) (point-max))))))
   (run-hooks 'nov-post-html-render-hook))
 
 (defun nov-render-document ()
@@ -539,14 +560,21 @@ the HTML is rendered with `nov-render-html-function'."
     (when done
       (1- i))))
 
+(defun nov-goto-document (index)
+  "Go to the document denoted by INDEX."
+  (let ((history (cons (list nov-documents-index (point))
+                       nov-history)))
+    (setq nov-documents-index index)
+    (nov-render-document)
+    (setq nov-history history)))
+
 (defun nov-goto-toc ()
   "Go to the TOC index and render the TOC document."
   (interactive)
   (let ((index (nov-find-document (lambda (doc) (eq (car doc) nov-toc-id)))))
     (when (not index)
       (error "Couldn't locate TOC"))
-    (setq nov-documents-index index)
-    (nov-render-document)))
+    (nov-goto-document index)))
 
 (defun nov-view-source ()
   "View the source of the current document in a new buffer."
@@ -593,15 +621,13 @@ the HTML is rendered with `nov-render-html-function'."
   "Go to the next document and render it."
   (interactive)
   (when (< nov-documents-index (1- (length nov-documents)))
-    (setq nov-documents-index (1+ nov-documents-index))
-    (nov-render-document)))
+    (nov-goto-document (1+ nov-documents-index))))
 
 (defun nov-previous-document ()
   "Go to the previous document and render it."
   (interactive)
   (when (> nov-documents-index 0)
-    (setq nov-documents-index (1- nov-documents-index))
-    (nov-render-document)))
+    (nov-goto-document (1- nov-documents-index))))
 
 (defun nov-scroll-up (arg)
   "Scroll with `scroll-up' or visit next chapter if at bottom."
@@ -622,22 +648,23 @@ the HTML is rendered with `nov-render-html-function'."
 
 (defun nov-visit-relative-file (filename target)
   "Visit the document as specified by FILENAME and TARGET."
-  (let* ((current-path (cdr (aref nov-documents nov-documents-index)))
-         (directory (file-name-directory current-path))
-         (path (file-truename (nov-make-path directory filename)))
-         (index (nov-find-document
-                 (lambda (doc) (equal path (file-truename (cdr doc)))))))
-    (when (not index)
-      (error "Couldn't locate document"))
-    (setq nov-documents-index index)
-    (let ((shr-target-id target))
-      (nov-render-document))
-    (when target
-      (let ((pos (next-single-property-change (point-min) 'shr-target-id)))
-        (when (not pos)
-          (error "Couldn't locate target"))
-        (goto-char pos)
-        (recenter (1- (max 1 scroll-margin)))))))
+  (when (not (zerop (length filename)))
+    (let* ((current-path (cdr (aref nov-documents nov-documents-index)))
+           (directory (file-name-directory current-path))
+           (path (file-truename (nov-make-path directory filename)))
+           (index (nov-find-document
+                   (lambda (doc) (equal path (file-truename (cdr doc)))))))
+      (when (not index)
+        (error "Couldn't locate document"))
+      (setq nov-documents-index index)))
+  (let ((shr-target-id target))
+    (nov-goto-document nov-documents-index))
+  (when target
+    (let ((pos (next-single-property-change (point-min) 'shr-target-id)))
+      (when (not pos)
+        (error "Couldn't locate target"))
+      (goto-char pos)
+      (recenter (1- (max 1 scroll-margin))))))
 
 ;; adapted from `shr-browse-url'
 (defun nov-browse-url (&optional mouse-event)
@@ -652,6 +679,15 @@ Internal URLs are visited with `nov-visit-relative-file'."
         (browse-url url)
       (apply 'nov-visit-relative-file (nov-url-filename-and-target url)))))
 
+(defun nov-copy-url (&optional mouse-event)
+  (interactive (list last-nonmenu-event))
+  (mouse-set-point mouse-event)
+  (let ((url (get-text-property (point) 'shr-url)))
+    (when (not url)
+      (user-error "No link under point"))
+    (kill-new url)
+    (message "%s" url)))
+
 (defun nov-saved-places ()
   "Retrieve saved places in `nov-save-place-file'."
   (when (and nov-save-place-file (file-exists-p nov-save-place-file))
@@ -681,6 +717,34 @@ Saving is only done if `nov-save-place-file' is set."
        (>= index 0)
        (< index (length documents))))
 
+(defun nov-history-back ()
+  "Go back in the history to the last visited document."
+  (interactive)
+  (or nov-history
+      (user-error "This is the first document you looked at"))
+  (let ((history-forward
+         (cons (list nov-documents-index (point))
+               nov-history-forward))
+        (index (car (car nov-history)))
+        (opoint (cadr (car nov-history))))
+    (setq nov-history (cdr nov-history))
+    (nov-goto-document index)
+    (setq nov-history (cdr nov-history))
+    (setq nov-history-forward history-forward)
+    (goto-char opoint)))
+
+(defun nov-history-forward ()
+  "Go forward in the history of visited documents."
+  (interactive)
+  (or nov-history-forward
+      (user-error "This is the last document you looked at"))
+  (let ((history-forward (cdr nov-history-forward))
+        (index (car (car nov-history-forward)))
+        (opoint (cadr (car nov-history-forward))))
+    (nov-goto-document index)
+    (setq nov-history-forward history-forward)
+    (goto-char opoint)))
+
 ;;;###autoload
 (define-derived-mode nov-mode special-mode "EPUB"
   "Major mode for reading EPUB documents"
@@ -713,6 +777,8 @@ Saving is only done if `nov-save-place-file' is set."
     (setq nov-documents-index 0))
   (setq buffer-undo-list t)
   (setq nov-file-name (buffer-file-name))
+  (setq-local bookmark-make-record-function
+              'nov-bookmark-make-record)
   (set-visited-file-name nil t) ; disable autosaves and save questions
   (let ((place (nov-saved-place (cdr (assq 'identifier nov-metadata)))))
     (if place
@@ -728,9 +794,8 @@ Saving is only done if `nov-save-place-file' is set."
       (nov-render-document))))
 
 
-;;; interop
+;;; recentf interop
 
-(require 'recentf)
 (defun nov-add-to-recentf ()
   (when nov-file-name
     (recentf-add-file nov-file-name)))
@@ -738,5 +803,61 @@ Saving is only done if `nov-save-place-file' is set."
 (add-hook 'nov-mode-hook 'nov-add-to-recentf)
 (add-hook 'nov-mode-hook 'hack-dir-local-variables-non-file-buffer)
 
+
+(defun nov--find-file (file index point)
+  "Open FILE(nil means current buffer) in nov-mode and go to the specified INDEX and POSITION."
+  (when file
+    (find-file file))
+  (unless (eq major-mode 'nov-mode)
+    (nov-mode))
+  (when (not (nov--index-valid-p nov-documents index))
+    (error "Invalid documents index"))
+  (setq nov-documents-index index)
+  (nov-render-document)
+  (goto-char point))
+
+;; Bookmark Integration
+(defun nov-bookmark-make-record  ()
+  "Create a bookmark epub record."
+  (cons (buffer-name)
+        `((filename . ,nov-file-name)
+          (index . ,nov-documents-index)
+          (position . ,(point))
+          (handler . nov-bookmark-jump-handler))))
+
+(defun nov-bookmark-jump-handler (bmk)
+  "The bookmark handler-function interface for bookmark BMK.
+
+See also `nov-bookmark-make-record'."
+  (let ((file (bookmark-prop-get bmk 'filename))
+        (index (bookmark-prop-get bmk 'index))
+        (position (bookmark-prop-get bmk 'position)))
+    (nov--find-file file index position)))
+
+
+;;; org interop
+
+(defun nov-org-link-follow (path)
+  (if (string-match "^\\(.*\\)::\\([0-9]+\\):\\([0-9]+\\)$" path)
+      (let ((file (match-string 1 path))
+            (index (string-to-number (match-string 2 path)))
+            (point (string-to-number (match-string 3 path))))
+        (nov--find-file file index point))
+    (error "Invalid nov.el link")))
+
+(defun nov-org-link-store ()
+  (when (and (eq major-mode 'nov-mode) nov-file-name)
+    (when (not (integerp nov-documents-index))
+      (setq nov-documents-index 0))
+    (org-store-link-props
+     :type "nov"
+     :link (format "nov:%s::%d:%d" nov-file-name nov-documents-index (point))
+     :description (format "EPUB file at %s" nov-file-name))))
+
+(org-link-set-parameters
+ "nov"
+ :follow 'nov-org-link-follow
+ :store 'nov-org-link-store)
+
 (provide 'nov)
 ;;; nov.el ends here
